from xaal.schemas import devices
from xaal.lib import tools,Device

import ujson
import logging
import socket
import binascii

from Crypto.Cipher import AES

GW_PORT = 9898
AQARA_ENCRYPT_IV = b'\x17\x99\x6d\x09\x3d\x28\xdd\xb3\xba\x69\x5a\x2e\x6f\x58\x56\x2e'

logger = logging.getLogger(__name__)

class AqaraDev(object):
    def __init__(self,sid,model,base_addr,xaal_gw):
        self.sid = sid
        self.model = model
        self.base_addr = base_addr
        self.xaal_gw = xaal_gw
        # xAAL embeded devices
        self.devices = [] 
        logger.info('New AqaraDevice %s %s' % (model,sid))
        self.setup()
        self.init_properties()
  
    def setup(self):
        logger.warning('Please overide setup()')

    def init_properties(self):
        for dev in self.devices:
            dev.vendor_id = 'Xiaomi / Aqara'
            dev.product_id = self.model
            dev.hw_id = self.sid
            dev.info = 'Aqara device: %s / %s' % (self.model,self.sid)
            if len(self.devices) > 1:
                dev.group_id = self.base_addr + 'ff'

    def parse(self,pkt):
        cmd = pkt.get('cmd',None)
        if not cmd:
            logger.warning('pkt w/ no command: %s' % pkt)
            return
        if cmd in ['report','heartbeat']:
            pload = pkt.get('data',None)
            if pload:
                # json in json really ? grr
                data = ujson.decode(pload)
                if cmd == 'report':
                    self.on_report(data)
                if cmd == 'heartbeat':
                    self.on_heartbeat(data)
        else:
            logger.info(pkt)

    def on_report(self,data):
        logger.info('Unhandled report %s' % data)

    def on_heartbeat(self,data):
        logger.info('Unhandled heartbeat %s' % data)


class Switch(AqaraDev):
    def setup(self):
        dev = Device('button.basic',self.base_addr+'00')
        self.devices.append(dev)

    def on_report(self,data):
        status = data.get('status',None)
        if status:
            self.devices[0].send_notification(status)


class Switch86sw2(AqaraDev):
    def setup(self):
        btn1 = Device('button.basic',self.base_addr+'00')
        btn2 = Device('button.basic',self.base_addr+'01')
        btn3 = Device('button.basic',self.base_addr+'02')
        self.devices = self.devices + [btn1,btn2,btn3]

    def on_report(self,data):
        chans = ['channel_0','channel_1','dual_channel']
        idx = 0 
        for k in chans:
            r = data.get(k,None)
            # in the current firmware this switch alsway return both click and long_click
            if r and (r.startswith('long_')==False):
                if r=='both_click': r='click'
                self.devices[idx].send_notification(r)
            idx = idx + 1

class Weather(AqaraDev):
    def setup(self):
        self.devices.append(devices.thermometer(self.base_addr+'00'))
        self.devices.append(devices.hygrometer(self.base_addr+'01'))
        self.devices.append(devices.barometer(self.base_addr+'02'))

    def on_report(self,data):
        val = data.get('temperature',None)
        if val: self.devices[0].attributes['temperature'] = round(int(val) / 100.0,1)
        val = data.get('humidity',None)
        if val: self.devices[1].attributes['humidity'] = round(int(val) / 100.0,1)
        val = data.get('pressure',None)
        if val: self.devices[2].attributes['pressure'] = round(int(val) / 100.0,1)

class Motion(AqaraDev):
    def setup(self):
        self.devices.append(devices.motion(self.base_addr+'00'))

    def on_report(self,data):
        val = data.get('status',None)
        if val and val == 'motion':
            self.devices[0].attributes['presence'] = True
        val = data.get('no_motion',None)
        if val:
            self.devices[0].attributes['presence'] = False


class Magnet(AqaraDev):
    def setup(self):
        dev = devices.door(self.base_addr+'00')
        dev.unsupported_methods = ['open','close']
        self.devices.append(dev)

    def on_report(self,data):
        status = data.get('status',None)
        if status and status == 'open':
            self.devices[0].attributes['position'] = True
        if status and status == 'close':
            self.devices[0].attributes['position'] = False


class Vibration(AqaraDev):pass


class Cube(AqaraDev):pass

class RelayController(AqaraDev):pass

class WaterLeak(AqaraDev):
    # no device yet on this

    def on_report(self,data):
        status = data.get('status',None)
        if status and status == 'leak':
            logger.warning('Leaking..')
        if status and status == 'no_leak':
            logger.warning('No leaking')


class Gateway(AqaraDev):
    def setup(self):
        self.sids = []
        self.ip = None
        self.token = None
        self.ready = False
        self.rgb = None
        self._blink_color = 0
        self.connect()
        lamp = devices.lamp(self.base_addr+'00')
        lamp.methods['on'] = self.lamp_on
        lamp.methods['off'] = self.lamp_off
        lamp.methods['rand']   = self.lamp_rand
        lamp.methods['blink']   = self.lamp_blink
        self.devices.append(lamp)

        siren = Device('siren.sound',self.base_addr+'01')
        siren.methods['play'] = self.siren_play
        siren.methods['stop'] = self.siren_stop
        #siren.methods['debug'] = self.debug
        self.devices.append(siren)

    def debug(self):
        import pdb;pdb.set_trace()

    #========================================================================
    ## GW Unicast methods
    #========================================================================  
    @property
    def secret(self):
        return self._secret

    @secret.setter
    def secret(self,value):
        logger.debug('Seting GW secret to %s' % value)
        self._secret = value

    def make_key(self):
        cipher = AES.new(self.secret, AES.MODE_CBC, IV=AQARA_ENCRYPT_IV)  
        return binascii.hexlify(cipher.encrypt(self.token)).decode("utf-8")

    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(0.5)

    def send(self,pkt):
        if not self.ip:
            logger.warning("GW IP not found yet, please wait")
            return
        try:
            #print(pkt)
            self.sock.sendto(pkt,(self.ip,GW_PORT))
            ans = self.sock.recv(65507)
            return ans
        except Exception as e:
            logger.warning(e)

    def send_cmd(self,cmd,sid,data={}):
        if not self.token :
            logger.warning("No token yet, please wait")
            return
        if not self.secret:
            logger.warn("Please set the secret key in cfg file")
            return
        key = self.make_key()
        data.update({"key":key})
        pload = {"cmd":   cmd,
                 "sid":   sid,
                 "data":  data
                }
        pkt = ujson.encode(pload).encode('utf8')
        return self.send(pkt)


    #========================================================================
    ## Gateway Unicast commands
    #========================================================================  
    def write(self,data = {}):
        """send command to the GW, used for siren, lamp, radio?"""
        pkt = self.send_cmd("write",self.sid,data)
        self.on_receive(pkt)

    def get_id_list(self):
        """ retreive the devices list"""
        pkt = self.send_cmd("get_id_list",self.sid)
        self.on_receive(pkt)

    def discover(self):
        """
        query device list and ask for a read, on each sid

        The doc say we should use {"cmd": "discovery"} instead of get_id_list + 
        read, but I always have a discory error => "missing sid"
        http://docs.opencloud.aqara.com/en/development/gateway-LAN-communication/
        """
        self.get_id_list()
        for sid in self.sids:
            pkt = self.send_cmd("read",sid)
            self.on_receive(pkt)

    def on_receive(self,pkt):         
        if not pkt: return 

        pload = ujson.decode(pkt)
        cmd = pload.get('cmd',None)

        if cmd == 'get_id_list_ack':
            data = pload.get('data','[]')
            self.sids = ujson.decode(data)

        if cmd == 'read_ack':
            model = pload.get('model',None)
            sid = pload.get('sid',None)
            dev = self.xaal_gw.get_device(sid,model)
            if dev:
                data = ujson.decode(pload.get('data','{}'))
                dev.on_report(data)
            else:
                logger.warning("Unknow device %s" % pload)

        if cmd == 'write_ack':
            print(pkt)


    #========================================================================
    ## RGB Leds 
    #========================================================================
    def get_rgb(self,red,green,blue,brightness=0xFF):
        return brightness<<24|( red << 16)|( green << 8)|blue

    def lamp_set(self,value):
        data = {"rgb" : value}
        self.write(data)

    def lamp_on(self):
        color = self.rgb or self.get_rgb(255,170,170)
        self.lamp_set(color)

    def lamp_off(self):
        self.lamp_set(0)

    def lamp_rand(self):
        import random
        red   = random.randint(0,255)
        green = random.randint(0,255)
        blue  = random.randint(0,255)
        rgb = self.get_rgb(red,green,blue)
        self.lamp_set(rgb)

    def _red(self):
        self.lamp_set(self.get_rgb(255,0,0))

    def _white(self):
        self.lamp_set(self.get_rgb(255,255,255))
        
    def _lamp_blink(self):
        if self._blink_color ==0:
            self._red()
            self._blink_color = 1
        else:
            self._white()
            self._blink_color = 0

    def lamp_blink(self):
        eng = self.devices[0].engine
        eng.add_timer(self.lamp_off,20,0)
        eng.add_timer(self._lamp_blink,1,15)
            
    #========================================================================
    ## Siren
    #========================================================================  
    def siren_play(self,_sound=2,_volume=5):
        logger.info('Playing %s %s' % (_sound,_volume))
        data = {
                "mid" : int(_sound),
                "volume": int(_volume),
                }
        self.write(data)

    def siren_stop(self):
        data = { "mid" : 999 }
        self.write(data)


    #========================================================================
    ## GW Mutlicast messages handlers
    #========================================================================  
    def parse(self,pkt):
        # extract the token before normal parsing
        token = pkt.get('token',None)
        if token: 
            self.token = token
        AqaraDev.parse(self,pkt)

    def on_heartbeat(self,data):
        ip = data.get('ip',None)
        if ip and ip != self.ip:
            self.ip = ip
            logger.info("GW IP found: %s" % ip)
        
        if not self.ready :
            if self.token and self.ip:
                self.ready = True
                self.discover()


    def on_report(self,data):
        rgb = data.get('rgb',None)
        if rgb:
            rgb = int(rgb)
            if rgb != 0:
                self.rgb = rgb
